@@ -33,8 +33,6 @@
 #include "libcman.h"
 #include "debug.h"
 
-#define LIBVIRT_XEN_URI "xen:///"
-
 static int running = 1;
 static int reload_key;
 
@@ -205,7 +203,7 @@ do_fence_request_tcp(fence_req_t *req, fence_auth_type_t auth,
 		     int flags)
 {
 	int fd = -1, ret = -1;
-	virDomainPtr vdp;
+	virDomainPtr vdp = NULL, nvdp = NULL;
 	virDomainInfo vdi;
 	char response = 1;
 	char *domain_desc, *domain_desc_sanitized;
@@ -305,9 +303,22 @@ do_fence_request_tcp(fence_req_t *req, fence_auth_type_t auth,
 			printf("Domain still exists; fencing failed\n");
 		} else if (domain_desc) {
 			/* Recreate the domain if possible */
-			/* Success */
+			/* Success... or not? */
 			dbg_printf(2, "Calling virDomainCreateLinux()...\n");
-			virDomainCreateLinux(vp, domain_desc, 0);
+			nvdp = virDomainCreateLinux(vp, domain_desc, 0);
+
+			if (nvdp == NULL) {
+				/* More recent versions of libvirt or perhaps the
+ 				   KVM back-end do not let you create a domain from
+ 				   XML if there is already a defined domain description
+ 				   with the same name that it knows about.  You must
+ 				   then call virDomainCreate() */
+				dbg_printf(2, "Failed; Trying virDomainCreate()...\n");
+				if (virDomainCreate(vdp) < 0) {
+					printf("Failed to recreate guest %s!\n",
+					       (char *)req->domain);
+				}
+			}
 			free(domain_desc);
 		}
 		break;
@@ -318,6 +329,10 @@ do_fence_request_tcp(fence_req_t *req, fence_auth_type_t auth,
 		perror("write");
 	}
 out:
+	if (vdp)
+		virDomainFree(vdp);
+	if (nvdp)
+		virDomainFree(nvdp);
 	if (fd != -1)
 		close(fd);
 
@@ -548,7 +563,7 @@ xvmd_loop(cman_handle_t ch, void *h, int fd, fence_xvm_args_t *args,
 	virt_list_t *vl = NULL;
 	virt_state_t *dom = NULL;
 
-	vp = virConnectOpen(LIBVIRT_XEN_URI);
+	vp = virConnectOpen(args->uri);
 	if (!vp)
 		perror("virConnectOpen");
 
@@ -605,7 +620,7 @@ xvmd_loop(cman_handle_t ch, void *h, int fd, fence_xvm_args_t *args,
 			continue;
 	
 		/* Request and/or timeout: open connection */
-		vp = virConnectOpen(LIBVIRT_XEN_URI);
+		vp = virConnectOpen(args->uri);
 		if (!vp) {
 			printf("NOTICE: virConnectOpen(): %s; cannot fence!\n",
 			       strerror(errno));
@@ -720,7 +735,7 @@ main(int argc, char **argv)
 	int mc_sock;
 	char key[MAX_KEY_LEN];
 	int key_len = 0, x;
-	char *my_options = "dfi:a:p:I:C:c:k:u?hLXV";
+	char *my_options = "dfi:a:p:I:C:U:c:k:u?hLXV";
 	cman_handle_t ch = NULL;
 	void *h = NULL;
 
@@ -738,12 +753,17 @@ main(int argc, char **argv)
 	}
 
 	if (args.flags & F_ERR) {
-		args_usage(argv[0], my_options, 0);
 		return 1;
 	}
 
 	if (args.flags & F_HELP) {
 		args_usage(argv[0], my_options, 0);
+
+		printf("Arguments may be specified as part of the\n");
+		printf("fence_xvmd tag in cluster.conf in the form of:\n");
+		printf("    <fence_xvmd argname=\"value\" ... />\n\n");
+
+		args_usage(argv[0], my_options, 1);
 		return 0;
 	}
 
