@@ -1,21 +1,8 @@
-/******************************************************************************
-*******************************************************************************
-**
-**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
-**
-**  This copyrighted material is made available to anyone wishing to use,
-**  modify, copy, or redistribute it subject to the terms and conditions
-**  of the GNU General Public License v.2.
-**
-*******************************************************************************
-******************************************************************************/
-
 #include <linux/sched.h>
 #include <linux/slab.h>
 #include <linux/smp_lock.h>
 #include <linux/spinlock.h>
-#include <asm/semaphore.h>
+#include <linux/semaphore.h>
 #include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <asm/uaccess.h>
@@ -23,9 +10,12 @@
 #include <linux/uio.h>
 #include <linux/blkdev.h>
 #include <linux/mm.h>
+#include <linux/aio.h>
+#include <linux/writeback.h>
 #include <asm/uaccess.h>
-#include <linux/gfs_ioctl.h>
+#include <linux/compat.h>
 
+#include "gfs_ioctl.h"
 #include "gfs.h"
 #include "bmap.h"
 #include "dio.h"
@@ -74,10 +64,10 @@ struct filldir_reg {
 };
 
 typedef ssize_t(*do_rw_t) (struct file * file,
-						   char *buf,
-						   size_t size, loff_t * offset,
+			   char *buf,
+			   size_t size, loff_t * offset,
                            struct kiocb *iocb,
-						   unsigned int num_gh, struct gfs_holder * ghs);
+			   unsigned int num_gh, struct gfs_holder * ghs);
 
 /**
  * gfs_llseek - seek to a location in a file
@@ -103,11 +93,11 @@ gfs_llseek(struct file *file, loff_t offset, int origin)
 	if (origin == 2) {
 		error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
 		if (!error) {
-			error = remote_llseek(file, offset, origin);
+			error = generic_file_llseek_unlocked(file, offset, origin);
 			gfs_glock_dq_uninit(&i_gh);
 		}
 	} else
-		error = remote_llseek(file, offset, origin);
+		error = generic_file_llseek_unlocked(file, offset, origin);
 
 	return error;
 }
@@ -130,7 +120,7 @@ gfs_llseek(struct file *file, loff_t offset, int origin)
 
 static ssize_t
 walk_vm_hard(struct file *file, char *buf, size_t size, loff_t *offset,
-	     struct kiocb *iocb, do_rw_t operation)
+             struct kiocb *iocb, do_rw_t operation)
 {
 	struct gfs_holder *ghs;
 	unsigned int num_gh = 0;
@@ -205,7 +195,8 @@ walk_vm_hard(struct file *file, char *buf, size_t size, loff_t *offset,
 
 static ssize_t
 walk_vm(struct file *file, char *buf, size_t size, loff_t *offset,
-		struct kiocb *iocb,	do_rw_t operation)
+	struct kiocb *iocb,	
+	do_rw_t operation)
 {
 	if (current->mm) {
 		struct super_block *sb = file->f_dentry->d_inode->i_sb;
@@ -236,7 +227,7 @@ walk_vm(struct file *file, char *buf, size_t size, loff_t *offset,
 	}
 
  do_locks:
-	return walk_vm_hard(file, buf, size, offset,  iocb, operation);
+	return walk_vm_hard(file, buf, size, offset, iocb, operation);
 }
 
 /**
@@ -252,7 +243,7 @@ walk_vm(struct file *file, char *buf, size_t size, loff_t *offset,
 
 static ssize_t
 do_read_readi(struct file *file, char *buf, size_t size, loff_t *offset,
-			  struct kiocb *iocb)
+              struct kiocb *iocb)
 {
 	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
 	ssize_t count = 0;
@@ -269,6 +260,8 @@ do_read_readi(struct file *file, char *buf, size_t size, loff_t *offset,
 			size = 0x7FFFFFFFull - *offset;
 	}
 
+	/* ToDo: not sure about iocb .. wcheng
+	 */
 	count = gfs_readi(ip, buf, *offset, size, gfs_copy2user);
 
 	if (count > 0)
@@ -277,175 +270,6 @@ do_read_readi(struct file *file, char *buf, size_t size, loff_t *offset,
 	return count;
 }
 
-/**
- * do_read_direct - Read bytes from a file
- * @file: The file to read from
- * @buf: The buffer to copy into
- * @size: The amount of data requested
- * @offset: The current file offset
- * @num_gh: The number of other locks we need to do the read
- * @ghs: the locks we need plus one for our lock
- *
- * Outputs: Offset - updated according to number of bytes read
- *
- * Returns: The number of bytes read, errno on failure
- */
-
-static ssize_t
-do_read_direct(struct file *file, char *buf, size_t size, loff_t *offset,
-			   struct kiocb *iocb,
-			   unsigned int num_gh, struct gfs_holder *ghs)
-{
-	struct inode *inode = file->f_mapping->host;
-	struct gfs_inode *ip = get_v2ip(inode);
-	unsigned int state = LM_ST_DEFERRED;
-	int flags = 0;
-	unsigned int x;
-	ssize_t count = 0;
-	int error;
-
-	for (x = 0; x < num_gh; x++)
-		if (ghs[x].gh_gl == ip->i_gl) {
-			state = LM_ST_SHARED;
-			flags |= GL_LOCAL_EXCL;
-			break;
-		}
-
-	gfs_holder_init(ip->i_gl, state, flags, &ghs[num_gh]);
-
-	error = gfs_glock_nq_m(num_gh + 1, ghs);
-	if (error)
-		goto out;
-
-	error = -EINVAL;
-	if (gfs_is_jdata(ip))
-		goto out_gunlock;
-
-	if (gfs_is_stuffed(ip)) {
-		size_t mask = bdev_hardsect_size(inode->i_sb->s_bdev) - 1;
-
-		if (((*offset) & mask) || (((unsigned long)buf) & mask))
-			goto out_gunlock;
-
-		count = do_read_readi(file, buf, size & ~mask, offset, iocb);
-	}
-	else {
-		if (!iocb)
-			count = do_sync_read(file, buf, size, offset);
-		else {
-			struct iovec local_iov = { .iov_base = buf, .iov_len = size};
-
-			count = generic_file_aio_read(iocb, &local_iov, 1, *offset);
-			iocb->ki_pos = *offset;
-		}
-	}
-
-	error = 0;
-
- out_gunlock:
-	gfs_glock_dq_m(num_gh + 1, ghs);
-
- out:
-	gfs_holder_uninit(&ghs[num_gh]);
-
-	return (count) ? count : error;
-}
-
-/**
- * do_read_buf - Read bytes from a file
- * @file: The file to read from
- * @buf: The buffer to copy into
- * @size: The amount of data requested
- * @offset: The current file offset
- * @num_gh: The number of other locks we need to do the read
- * @ghs: the locks we need plus one for our lock
- *
- * Outputs: Offset - updated according to number of bytes read
- *
- * Returns: The number of bytes read, errno on failure
- */
-
-static ssize_t
-do_read_buf(struct file *file, char *buf, size_t size, loff_t *offset,
-			struct kiocb *iocb,
-			unsigned int num_gh, struct gfs_holder *ghs)
-{
-	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
-	ssize_t count = 0;
-	int error;
-
-	gfs_holder_init(ip->i_gl, LM_ST_SHARED, GL_ATIME, &ghs[num_gh]);
-
-	error = gfs_glock_nq_m_atime(num_gh + 1, ghs);
-	if (error)
-		goto out;
-
-	if (gfs_is_jdata(ip) ||
-	    (gfs_is_stuffed(ip) && !test_bit(GIF_PAGED, &ip->i_flags)))
-		count = do_read_readi(file, buf, size, offset, iocb);
-	else {
-		if (!iocb) {
-			count = do_sync_read(file, buf, size, offset);
-		} else {
-			struct iovec local_iov = { .iov_base = buf, .iov_len = size};
-
-			count = generic_file_aio_read(iocb, &local_iov, 1, *offset);
-			iocb->ki_pos = *offset;
-		}
-	}
-
-	gfs_glock_dq_m(num_gh + 1, ghs);
-
- out:
-	gfs_holder_uninit(&ghs[num_gh]);
-
-	return (count) ? count : error;
-}
-
-static ssize_t
-__gfs_read(struct file *file, char *buf, size_t size, loff_t *offset,
-		 struct kiocb *iocb)
-{
-	atomic_inc(&get_v2sdp(file->f_mapping->host->i_sb)->sd_ops_file);
-
-	if (file->f_flags & O_DIRECT)
-		return walk_vm(file, buf, size, offset, iocb, do_read_direct);
-	else
-		return walk_vm(file, buf, size, offset, iocb, do_read_buf);
-}
-
-/**
- * gfs_read - Read bytes from a file
- * @file: The file to read from
- * @buf: The buffer to copy into
- * @size: The amount of data requested
- * @offset: The current file offset
- *
- * Outputs: Offset - updated according to number of bytes read
- *
- * Returns: The number of bytes read, errno on failure
- */
-
-static ssize_t
-gfs_read(struct file *file, char *buf, size_t size, loff_t *offset)
-{
-	return(__gfs_read(file, buf, size, offset, NULL));
-}
-
-/*
- * gfs_aio_read: match with vfs generic_file_aio_read as:
- * 	(struct kiocb *iocb, char __user *buf, size_t count, loff_t pos)
- */
-static ssize_t
-gfs_aio_read(struct kiocb *iocb, const struct iovec *iov, unsigned long count,
-			 loff_t pos)
-{
-	struct file *filp = iocb->ki_filp;
-
-	BUG_ON(iocb->ki_pos != pos);
-	return(__gfs_read(filp, iov->iov_base, iov->iov_len, &iocb->ki_pos, iocb));
-}
-
 /**
  * grope_mapping - feel up a mapping that needs to be written
  * @buf: the start of the memory to be written
@@ -477,6 +301,186 @@ grope_mapping(char *buf, size_t size)
 	return 0;
 }
 
+/**
+ * do_read_direct - Read bytes from a file
+ * @file: The file to read from
+ * @buf: The buffer to copy into
+ * @size: The amount of data requested
+ * @offset: The current file offset
+ * @num_gh: The number of other locks we need to do the read
+ * @ghs: the locks we need plus one for our lock
+ *
+ * Outputs: Offset - updated according to number of bytes read
+ *
+ * Returns: The number of bytes read, errno on failure
+ */
+
+static ssize_t
+do_read_direct(struct file *file, char *buf, size_t size, loff_t *offset,
+		struct kiocb *iocb,
+		unsigned int num_gh, struct gfs_holder *ghs)
+{
+	struct inode *inode = file->f_mapping->host;
+	struct gfs_inode *ip = get_v2ip(inode);
+	unsigned int state = LM_ST_DEFERRED;
+	int flags = 0;
+	unsigned int x;
+	ssize_t count = 0;
+	int error;
+
+	for (x = 0; x < num_gh; x++)
+		if (ghs[x].gh_gl == ip->i_gl) {
+			state = LM_ST_SHARED;
+			flags |= GL_LOCAL_EXCL;
+			break;
+		}
+
+	gfs_holder_init(ip->i_gl, state, flags, &ghs[num_gh]);
+
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
+	error = gfs_glock_nq_m(num_gh + 1, ghs);
+	if (error)
+		goto out;
+
+	error = -EINVAL;
+	if (gfs_is_jdata(ip))
+		goto out_gunlock;
+
+	if (gfs_is_stuffed(ip)) {
+		size_t mask = bdev_hardsect_size(inode->i_sb->s_bdev) - 1;
+
+		if (((*offset) & mask) || (((unsigned long)buf) & mask))
+			goto out_gunlock;
+
+		count = do_read_readi(file, buf, size & ~mask, offset, iocb);
+        }
+        else {
+                if (!iocb) 
+                        count = do_sync_read(file, buf, size, offset);
+                else {
+                        struct iovec local_iov = { .iov_base = buf, .iov_len = size};
+
+                        count = generic_file_aio_read(iocb, &local_iov, 1, *offset);
+                        iocb->ki_pos = *offset;
+                }
+        }
+
+	error = 0;
+
+ out_gunlock:
+	gfs_glock_dq_m(num_gh + 1, ghs);
+
+ out:
+	gfs_holder_uninit(&ghs[num_gh]);
+
+	return (count) ? count : error;
+}
+
+/**
+ * do_read_buf - Read bytes from a file
+ * @file: The file to read from
+ * @buf: The buffer to copy into
+ * @size: The amount of data requested
+ * @offset: The current file offset
+ * @num_gh: The number of other locks we need to do the read
+ * @ghs: the locks we need plus one for our lock
+ *
+ * Outputs: Offset - updated according to number of bytes read
+ *
+ * Returns: The number of bytes read, errno on failure
+ */
+
+static ssize_t
+do_read_buf(struct file *file, char *buf, size_t size, loff_t *offset,
+		struct kiocb *iocb,
+		unsigned int num_gh, struct gfs_holder *ghs)
+{
+	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
+	ssize_t count = 0;
+	int error;
+
+	gfs_holder_init(ip->i_gl, LM_ST_SHARED, GL_ATIME, &ghs[num_gh]);
+
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
+	error = gfs_glock_nq_m_atime(num_gh + 1, ghs);
+	if (error)
+		goto out;
+
+	if (gfs_is_jdata(ip) ||
+	    (gfs_is_stuffed(ip) && !test_bit(GIF_PAGED, &ip->i_flags)))
+                count = do_read_readi(file, buf, size, offset, iocb);
+        else {
+                if (!iocb) {
+                        count = do_sync_read(file, buf, size, offset);
+                } else {
+                        struct iovec local_iov = { .iov_base = buf, .iov_len = size};
+
+                        count = generic_file_aio_read(iocb, &local_iov, 1, *offset);
+                        iocb->ki_pos = *offset;
+                }
+        }
+
+	gfs_glock_dq_m(num_gh + 1, ghs);
+
+ out:
+	gfs_holder_uninit(&ghs[num_gh]);
+
+	return (count) ? count : error;
+}
+
+static ssize_t
+__gfs_read(struct file *file, char *buf, size_t size, loff_t *offset, struct kiocb *iocb)
+{
+	atomic_inc(&get_v2sdp(file->f_mapping->host->i_sb)->sd_ops_file);
+
+	if (file->f_flags & O_DIRECT)
+		return walk_vm(file, buf, size, offset, iocb, do_read_direct);
+	else
+		return walk_vm(file, buf, size, offset, iocb, do_read_buf);
+}
+
+/**
+ * gfs_read - Read bytes from a file
+ * @file: The file to read from
+ * @buf: The buffer to copy into
+ * @size: The amount of data requested
+ * @offset: The current file offset
+ *
+ * Outputs: Offset - updated according to number of bytes read
+ *
+ * Returns: The number of bytes read, errno on failure
+ */
+
+static ssize_t
+gfs_read(struct file *file, char *buf, size_t size, loff_t *offset)
+{
+	return(__gfs_read(file, buf, size, offset, NULL));
+}
+
+/*
+ * gfs_aio_read: match with vfs generic_file_aio_read as:
+ *	(struct kiocb *iocb, char __user *buf, size_t count, loff_t pos)
+ */
+static ssize_t
+gfs_aio_read(struct kiocb *iocb, const struct iovec *iov, unsigned long count,
+		loff_t pos)
+{
+	struct file *filp = iocb->ki_filp;
+
+	BUG_ON(iocb->ki_pos != pos);
+	return(__gfs_read(filp, iov->iov_base, iov->iov_len, &iocb->ki_pos, iocb));
+}
+
 /**
  * gfs_file_aio_write_nolock - Call vfs aio layer to write bytes to a file
  * @file: The file to write to
@@ -489,27 +493,27 @@ grope_mapping(char *buf, size_t size)
  */
 static ssize_t
 gfs_file_aio_write_nolock(struct file *file, char *buf, size_t size,
-						  loff_t *offset, struct kiocb *iocb)
+                                                  loff_t *offset, struct kiocb *iocb)
 {
-	struct iovec local_iov = { .iov_base = buf, .iov_len = size };
-	struct kiocb local_iocb, *kiocb = NULL;
-	ssize_t count;
+        struct iovec local_iov = { .iov_base = buf, .iov_len = size };
+        struct kiocb local_iocb, *kiocb = NULL;
+        ssize_t count;
 
-	if (!iocb) {
-		init_sync_kiocb(&local_iocb, file);
-		local_iocb.ki_nr_segs = 1;
-		kiocb = &local_iocb;
-	}
-	else
-		kiocb = iocb;
-	
-	kiocb->ki_pos = *offset;
-	count = generic_file_aio_write_nolock(kiocb, &local_iov, kiocb->ki_nr_segs,
-										  kiocb->ki_pos);
-	*offset = kiocb->ki_pos;
-	if (kiocb == &local_iocb && count == -EIOCBQUEUED)
-		count = wait_on_sync_kiocb(kiocb);
-	return count;
+        if (!iocb) {
+                init_sync_kiocb(&local_iocb, file);
+                local_iocb.ki_nr_segs = 1;
+                kiocb = &local_iocb;
+        }
+        else
+                kiocb = iocb;
+        
+        kiocb->ki_pos = *offset;
+        count = generic_file_aio_write_nolock(kiocb, &local_iov, kiocb->ki_nr_segs,
+                                                                                  kiocb->ki_pos);
+        *offset = kiocb->ki_pos;
+        if (kiocb == &local_iocb && count == -EIOCBQUEUED)
+                count = wait_on_sync_kiocb(kiocb);
+        return count;
 }
 
 /**
@@ -578,9 +582,7 @@ do_write_direct_alloc(struct file *file, char *buf, size_t size, loff_t *offset,
 		brelse(dibh);
 	}
 
-	if (gfs_is_stuffed(ip)) {
-		error = gfs_unstuff_dinode(ip, gfs_unstuffer_sync, NULL);
-		if (error)
+	if (gfs_is_stuffed(ip)) { error = gfs_unstuff_dinode(ip, gfs_unstuffer_sync, NULL); if (error)
 			goto fail_end_trans;
 	}
 
@@ -604,6 +606,10 @@ do_write_direct_alloc(struct file *file, char *buf, size_t size, loff_t *offset,
 
 	gfs_trans_end(sdp);
 
+	/* Question (wcheng)
+	 * 1. should IS_SYNC flush glock ?
+	 * 2. does gfs_log_flush_glock flush data ?
+	 */
 	if (file->f_flags & O_SYNC)
 		gfs_log_flush_glock(ip->i_gl);
 
@@ -677,6 +683,12 @@ do_write_direct(struct file *file, char *buf, size_t size, loff_t *offset,
  restart:
 	gfs_holder_init(ip->i_gl, state, 0, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -752,6 +764,9 @@ do_write_direct(struct file *file, char *buf, size_t size, loff_t *offset,
 		if (error)
 			goto out_gunlock;
 
+		/* Todo: It would be nice if init_sync_kiocb is exported.
+		 *  .. wcheng
+		 */
 		count = gfs_file_aio_write_nolock(file, buf, size, offset, iocb);
 		gfs_glock_dq_uninit(&t_gh);
 	}
@@ -782,7 +797,7 @@ out:
 
 static ssize_t
 do_do_write_buf(struct file *file, char *buf, size_t size, loff_t *offset,
-				struct kiocb *iocb)
+		struct kiocb *iocb)
 {
 	struct inode *inode = file->f_mapping->host;
 	struct gfs_inode *ip = get_v2ip(inode);
@@ -950,9 +965,9 @@ do_do_write_buf(struct file *file, char *buf, size_t size, loff_t *offset,
 
 static ssize_t
 do_write_buf(struct file *file,
-			 char *buf, size_t size, loff_t *offset,
-			 struct kiocb *iocb,
-			 unsigned int num_gh, struct gfs_holder *ghs)
+		char *buf, size_t size, loff_t *offset,
+		struct kiocb *iocb,
+		unsigned int num_gh, struct gfs_holder *ghs)
 {
 	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
 	struct gfs_sbd *sdp = ip->i_sbd;
@@ -962,6 +977,12 @@ do_write_buf(struct file *file,
 
 	gfs_holder_init(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[num_gh]);
 
+	if (num_gh && atomic_read(&current->mm->mm_users) > 1) {
+		error = grope_mapping(buf, size);
+		if (error)
+			goto out;
+	}
+
 	error = gfs_glock_nq_m(num_gh + 1, ghs);
 	if (error)
 		goto out;
@@ -1052,14 +1073,14 @@ gfs_write(struct file *file, const char *buf, size_t size, loff_t *offset)
 
 static ssize_t
 gfs_aio_write(struct kiocb *iocb, const struct iovec *iov, unsigned long segs,
-			  loff_t pos)
+			loff_t pos)
 {
 	struct file *file = iocb->ki_filp;
 
 	BUG_ON(iocb->ki_pos != pos);
 
-	return(__gfs_write(file, iov->iov_base, iov->iov_len, &iocb->ki_pos, 
-					   iocb));
+	return(__gfs_write(file, iov->iov_base, iov->iov_len, &iocb->ki_pos,
+					iocb));
 }
 
 /**
@@ -1135,6 +1156,29 @@ filldir_reg_func(void *opaque,
 	return 0;
 }
 
+static inline int
+high_stat_rate(struct gfs_inode *ip)
+{
+	struct timeval now, diff;
+	unsigned long milli = 0;
+	unsigned int rate = 0; /* stats per second in this dir */
+
+	do_gettimeofday(&now);
+	diff.tv_sec = now.tv_sec - ip->i_dir_stat_st.tv_sec;
+	diff.tv_usec = now.tv_usec - ip->i_dir_stat_st.tv_usec;
+	if (diff.tv_usec < 0) {
+		diff.tv_usec += 1000000;
+		diff.tv_sec--;
+	}
+	milli = (diff.tv_sec * 1000) + (diff.tv_usec / 1000);
+	if (milli)
+		rate = (ip->i_dir_stats * 1000) / milli;
+
+	if (rate > 10) /* More than 10 stats/sec */
+		return 1;
+	return 0;
+}
+
 /**
  * readdir_reg - Read directory entries from a directory
  * @file: The directory to read from
@@ -1154,10 +1198,14 @@ readdir_reg(struct file *file, void *dirent, filldir_t filldir)
 	int error;
 
 	fdr.fdr_sbd = dip->i_sbd;
-	fdr.fdr_prefetch = TRUE;
+	fdr.fdr_prefetch = high_stat_rate(dip) ? TRUE : FALSE;
 	fdr.fdr_filldir = filldir;
 	fdr.fdr_opaque = dirent;
 
+	/* reset stat counter and timestamp */
+	dip->i_dir_stats = 0;
+	do_gettimeofday(&dip->i_dir_stat_st);
+
 	gfs_holder_init(dip->i_gl, LM_ST_SHARED, GL_ATIME, &d_gh);
 	error = gfs_glock_nq_atime(&d_gh);
 	if (error) {
@@ -1364,6 +1412,41 @@ gfs_ioctl(struct inode *inode, struct file *file,
 	}
 }
 
+#ifdef CONFIG_COMPAT
+/**
+ * gfs_compat_ioctl - do an ioctl on a file - compatible between 32-64 bit
+ * @inode: the inode
+ * @file: the file pointer
+ * @cmd: the ioctl command
+ * @arg: the argument
+ *
+ * Returns: errno
+ */
+
+static long
+gfs_compat_ioctl(struct file *file, unsigned cmd, unsigned long arg)
+{
+	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
+
+	atomic_inc(&ip->i_sbd->sd_ops_file);
+
+	switch (cmd) {
+	case GFS_IOCTL_IDENTIFY: {
+                unsigned int x = GFS_MAGIC;
+                if (copy_to_user((unsigned int *)arg, &x, sizeof(unsigned int)))
+                        return -EFAULT;
+		return 0;
+        }
+
+	case GFS_IOCTL_SUPER:
+		return gfs_ioctl_i_compat(ip, arg);
+
+	default:
+		return -ENOTTY;
+	}
+}
+#endif
+
 /**
  * gfs_mmap - We don't support shared writable mappings right now
  * @file: The file to map
@@ -1520,6 +1603,7 @@ gfs_fsync(struct file *file, struct dentry *dentry, int datasync)
 {
 	struct gfs_inode *ip = get_v2ip(dentry->d_inode);
 	struct gfs_holder i_gh;
+	struct inode *inode = dentry->d_inode;
 	int error;
 
 	atomic_inc(&ip->i_sbd->sd_ops_file);
@@ -1530,8 +1614,15 @@ gfs_fsync(struct file *file, struct dentry *dentry, int datasync)
 
 	if (gfs_is_jdata(ip))
 		gfs_log_flush_glock(ip->i_gl);
-	else
-		i_gh.gh_flags |= GL_SYNC;
+	else {
+		if ((!datasync) || (inode->i_state & I_DIRTY_DATASYNC)) {
+			struct writeback_control wbc = {
+				.sync_mode = WB_SYNC_ALL,
+				.nr_to_write = 0,
+			};
+			error = sync_inode(inode, &wbc);
+		}
+	}
 
 	gfs_glock_dq_uninit(&i_gh);
 
@@ -1563,24 +1654,10 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
 	if ((ip->i_di.di_mode & (S_ISGID | S_IXGRP)) == S_ISGID)
 		return -ENOLCK;
 
-	if (sdp->sd_args.ar_localflocks) {
-		if (IS_GETLK(cmd)) {
-			struct file_lock tmp;
-			int ret;
-			lock_kernel();
-			ret = posix_test_lock(file, fl, &tmp);
-			fl->fl_type = F_UNLCK;
-			if (ret)
-				memcpy(fl, &tmp, sizeof(struct file_lock));
-			unlock_kernel();
-		        return 0;
-		} else {
-			int error;
-			lock_kernel();
-			error = posix_lock_file_wait(file, fl);
-			unlock_kernel();
-			return error;
-		}
+	if (cmd == F_CANCELLK) {
+		/* Hack: */
+		cmd = F_SETLK;
+		fl->fl_type = F_UNLCK;
 	}
 
 	if (IS_GETLK(cmd))
@@ -1592,11 +1669,11 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
 }
 
 /**
- * gfs_sendfile - Send bytes to a file or socket
+ * gfs_splice_read - Send bytes to a file or socket
  * @in_file: The file to read from
  * @out_file: The file to write to
  * @count: The amount of data
- * @offset: The beginning file offset
+ * @ppos: The beginning file offset
  *
  * Outputs: offset - updated according to number of bytes read
  *
@@ -1604,7 +1681,7 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
  */
 
 static ssize_t
-gfs_sendfile(struct file *in_file, loff_t *offset, size_t count, read_actor_t actor, void __user *target)
+gfs_splice_read(struct file *in_file, loff_t *ppos, struct pipe_inode_info *pipe, size_t count, unsigned int flags)
 {
 	struct gfs_inode *ip = get_v2ip(in_file->f_mapping->host);
 	struct gfs_holder gh;
@@ -1621,7 +1698,7 @@ gfs_sendfile(struct file *in_file, loff_t *offset, size_t count, read_actor_t ac
 	if (gfs_is_jdata(ip))
 		retval = -ENOSYS;
 	else 
-		retval = generic_file_sendfile(in_file, offset, count, actor, target);
+		retval = generic_file_splice_read(in_file, ppos, pipe, count, flags);
 
 	gfs_glock_dq(&gh);
 
@@ -1724,7 +1801,6 @@ static int
 gfs_flock(struct file *file, int cmd, struct file_lock *fl)
 {
 	struct gfs_inode *ip = get_v2ip(file->f_mapping->host);
-	struct gfs_sbd *sdp = ip->i_sbd;
 
 	atomic_inc(&ip->i_sbd->sd_ops_file);
 
@@ -1733,9 +1809,6 @@ gfs_flock(struct file *file, int cmd, struct file_lock *fl)
 	if ((ip->i_di.di_mode & (S_ISGID | S_IXGRP)) == S_ISGID)
 		return -ENOLCK;
 
-	if (sdp->sd_args.ar_localflocks)
-		return flock_lock_file_wait(file, fl);
-
 	if (fl->fl_type == F_UNLCK) {
 		do_unflock(file, fl);
 		return 0;
@@ -1750,21 +1823,55 @@ struct file_operations gfs_file_fops = {
 	.aio_read = gfs_aio_read,
 	.aio_write = gfs_aio_write,
 	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl   = gfs_compat_ioctl,
+#endif
 	.mmap = gfs_mmap,
 	.open = gfs_open,
 	.release = gfs_close,
 	.fsync = gfs_fsync,
 	.lock = gfs_lock,
-	.sendfile = gfs_sendfile,
+	.splice_read = gfs_splice_read,
 	.flock = gfs_flock,
 };
 
 struct file_operations gfs_dir_fops = {
 	.readdir = gfs_readdir,
 	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl   = gfs_compat_ioctl,
+#endif
 	.open = gfs_open,
 	.release = gfs_close,
 	.fsync = gfs_fsync,
 	.lock = gfs_lock,
 	.flock = gfs_flock,
 };
+
+struct file_operations gfs_file_fops_nolock = {
+	.llseek = gfs_llseek,
+	.read = gfs_read,
+	.write = gfs_write,
+	.aio_read = gfs_aio_read,
+	.aio_write = gfs_aio_write,
+	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl = gfs_compat_ioctl,
+#endif
+	.mmap = gfs_mmap,
+	.open = gfs_open,
+	.release = gfs_close,
+	.fsync = gfs_fsync,
+	.splice_read = gfs_splice_read,
+};
+
+struct file_operations gfs_dir_fops_nolock = {
+	.readdir = gfs_readdir,
+	.ioctl = gfs_ioctl,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl = gfs_compat_ioctl,
+#endif
+	.open = gfs_open,
+	.release = gfs_close,
+	.fsync = gfs_fsync,
+};
