@@ -296,31 +296,31 @@ static void are_you_sure(struct gfs2_sbd *sdp)
 }
 
 /**
- * check_mount - check to see if device is mounted
+ * check_mount - check to see if device is mounted/busy
  * @device: the device to create the filesystem on
  *
  */
 
 void check_mount(char *device)
 {
-	struct mntent *mnt;
-	FILE *fp;
+	struct stat st_buf;
+	int fd;
 
-	if ((fp = setmntent("/proc/mounts", "r")) == NULL) {
-		die("error opening /proc/mounts");
-	}
+	if (stat(device, &st_buf) < 0)
+		die("could not stat device %s\n", device);
+	if (!S_ISBLK(st_buf.st_mode))
+		die("%s is not a block device\n", device);
 
-	while ((mnt = getmntent(fp)) != NULL) {
-		if (strcmp(device, mnt->mnt_fsname) == 0) {
-			printf("cannot create filesystem: ");
-			printf("%s appears to be mounted\n", device);
-			endmntent(fp);
-			exit(EXIT_FAILURE);
-			break;
+	fd = open(device, O_RDONLY | O_NONBLOCK | O_EXCL);
+
+	if (fd < 0) {
+		if (errno == EBUSY) {
+			die("device %s is busy\n", device);
 		}
 	}
-
-	endmntent(fp);
+	else {
+		close(fd);
+	}
 
 	return;
 }
@@ -359,7 +359,8 @@ print_results(struct gfs2_sbd *sdp, uint64_t real_device_size)
 
 	if (sdp->debug) {
 		printf("\n");
-		printf("Spills:                    %u\n", sdp->spills);
+		printf("Spills:                    %u\n",
+		       sdp->buf_list.spills);
 		printf("Writes:                    %u\n", sdp->writes);
 	}
 
@@ -377,7 +378,6 @@ void
 main_mkfs(int argc, char *argv[])
 {
 	struct gfs2_sbd sbd, *sdp = &sbd;
-	unsigned int x;
 	int error;
 	int rgsize_specified = 0;
 	uint64_t real_device_size;
@@ -391,9 +391,8 @@ main_mkfs(int argc, char *argv[])
 	strcpy(sdp->lockproto, GFS2_DEFAULT_LOCKPROTO);
 	sdp->time = time(NULL);
 	osi_list_init(&sdp->rglist);
-	osi_list_init(&sdp->buf_list);
-	for (x = 0; x < BUF_HASH_SIZE; x++)
-		osi_list_init(&sdp->buf_hash[x]);
+	init_buf_list(sdp, &sdp->buf_list, 128 << 20);
+	init_buf_list(sdp, &sdp->nvbuf_list, 0xffffffff);
 
 	decode_arguments(argc, argv, sdp);
 	if (sdp->rgsize == -1)                 /* if rg size not specified */
@@ -459,7 +458,8 @@ main_mkfs(int argc, char *argv[])
 	inode_put(sdp->master_dir, updated);
 	inode_put(sdp->md.inum, updated);
 	inode_put(sdp->md.statfs, updated);
-	bsync(sdp);
+	bsync(&sdp->buf_list);
+	bsync(&sdp->nvbuf_list);
 
 	error = fsync(sdp->device_fd);
 	if (error)
