@@ -32,7 +32,7 @@ do_df_one(char *path)
 	uint64_t rgrps;
 	unsigned int flags;
 	char *value, *fs;
- 	int x, statfs_fd;
+	int statfs_fd;
 	struct gfs2_sbd sbd;
 	char buf[GFS2_DEFAULT_BSIZE], statfs_fn[PATH_MAX];
 	struct gfs2_statfs_change sc;
@@ -52,9 +52,9 @@ do_df_one(char *path)
 	sbd.utsize = GFS2_DEFAULT_UTSIZE;
 	sbd.qcsize = GFS2_DEFAULT_QCSIZE;
 	osi_list_init(&sbd.rglist);
-	osi_list_init(&sbd.buf_list);
-	for (x = 0; x < BUF_HASH_SIZE; x++)
-		osi_list_init(&sbd.buf_hash[x]);
+	init_buf_list(&sbd, &sbd.buf_list, 128 << 20);
+	init_buf_list(&sbd, &sbd.nvbuf_list, 0xffffffff);
+
 	do_lseek(sbd.device_fd, 0x10 * sbd.bsize);
 	do_read(sbd.device_fd, buf, sbd.bsize); /* read in the superblock */
 
@@ -101,9 +101,7 @@ do_df_one(char *path)
 		(get_sysfs_uint(fs, "args/localcaching")) ? "TRUE" : "FALSE");
 
 	/* Read the master statfs file */
-	if (!find_gfs2_meta(&sbd))
-		mount_gfs2_meta(&sbd);
-	lock_for_admin(&sbd);
+	mount_gfs2_meta(&sbd);
 
 	sprintf(statfs_fn, "%s/statfs", sbd.metafs_path);
 	statfs_fd = open(statfs_fn, O_RDONLY);
@@ -121,14 +119,17 @@ do_df_one(char *path)
 	percentage = sc.sc_total ?
 		(100.0 * (sc.sc_total - sc.sc_free)) / sc.sc_total + 0.5 : 0;
 	printf("  %-15s%-15llu%-15llu%-15llu%u%%\n", "data",
-	       sc.sc_total, sc.sc_total - sc.sc_free, sc.sc_free, percentage);
+	       (unsigned long long)sc.sc_total,
+	       (unsigned long long)sc.sc_total - sc.sc_free,
+	       (unsigned long long)sc.sc_free, percentage);
 
 	percentage = (sc.sc_dinodes + sc.sc_free) ?
 		(100.0 * sc.sc_dinodes / (sc.sc_dinodes + sc.sc_free)) + 0.5 :
 		0;
 	printf("  %-15s%-15llu%-15llu%-15llu%u%%\n", "inodes",
-	       sc.sc_dinodes + sc.sc_free, sc.sc_dinodes,
-	       sc.sc_free, percentage);
+	       (unsigned long long)sc.sc_dinodes + sc.sc_free,
+	       (unsigned long long)sc.sc_dinodes,
+	       (unsigned long long)sc.sc_free, percentage);
 }
 
 
