@@ -1,21 +1,3 @@
-/*
-  Copyright Red Hat, Inc. 2002-2003
-
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2, or (at your option) any
-  later version.
-
-  This program is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  General Public License for more details.
-
-  You should have received a copy of the GNU General Public License
-  along with this program; see the file COPYING.  If not, write to the
-  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge,
-  MA 02139, USA.
-*/
 /** @file
  * Fail-over Domain & Preferred Node Ordering Driver.  Ripped right from
  * the clumanager 1.2 code base.
@@ -67,7 +49,11 @@
 int group_property(char *, char *, char *, size_t);
 
 fod_node_t *
+#ifndef NO_CCS
 fod_get_node(int ccsfd, char *base, int idx, fod_t *domain)
+#else
+fod_get_node(int __attribute__((unused)) ccsfd, char *base, int idx, fod_t *domain)
+#endif
 {
 	fod_node_t *fodn;
 	char xpath[256];
@@ -365,13 +351,24 @@ node_in_domain(char *nodename, fod_t *domain,
 
 
 int
-node_domain_set(fod_t *domain, int **ret, int *retlen)
+node_domain_set(fod_t **domains, char *name, int **ret, int *retlen, int *flags)
 {
 	int x, i, j;
 	int *tmpset;
 	int ts_count;
-
 	fod_node_t *fodn;
+	fod_t *domain;
+	int found = 0;
+
+	list_for(domains, domain, x) {
+		if (!strcasecmp(domain->fd_name, name)) {
+			found = 1;
+			break;
+		}
+	} // while (!list_done(&_domains, fod));
+
+	if (!found)
+		return -1;
 
 	/* Count domain length */
 	list_for(&domain->fd_nodes, fodn, x) { }
@@ -384,6 +381,8 @@ node_domain_set(fod_t *domain, int **ret, int *retlen)
 	if (!(*tmpset))
 		return -1;
 
+	*flags = domain->fd_flags;
+
 	if (domain->fd_flags & FOD_ORDERED) {
 		for (i = 1; i <= 100; i++) {
 			
@@ -543,7 +542,7 @@ node_should_start(int nodeid, cluster_member_list_t *membership,
 		 */
 		if (svc_state.rs_state == RG_STATE_STARTED)
 			started = 1;
-		if (svc_state.rs_owner == nodeid)
+		if (svc_state.rs_owner == (uint32_t)nodeid)
 			owned_by_node = 1;
 		if (!memb_online(membership, svc_state.rs_owner))
 			no_owner = 1;
