@@ -1,24 +1,3 @@
-/*
-  Copyright Red Hat, Inc. 2004-2006
-
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2, or (at your option) any
-  later version.
-
-  This program is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  General Public License for more details.
-
-  You should have received a copy of the GNU General Public License
-  along with this program; see the file COPYING.  If not, write to the
-  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
-  MA 02139, USA.
-
-  Fix for #193859 - relocation of a service w/o umounting file-systems
-    by Navid Sheikhol-Eslami [ navid at redhat dot com ]
-*/
 #include <libxml/parser.h>
 #include <libxml/xmlmemory.h>
 #include <libxml/xpath.h>
@@ -30,6 +9,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <list.h>
+#include <restart_counter.h>
 #include <reslist.h>
 #include <pthread.h>
 #include <clulog.h>
@@ -129,7 +109,7 @@ kill_env(char **env)
    @see			build_env
  */
 static void
-add_ocf_stuff(resource_t *res, char **env, int depth)
+add_ocf_stuff(resource_t *res, char **env, int depth, int refcnt)
 {
 	char ver[10];
 	char *minor, *val;
@@ -213,6 +193,17 @@ add_ocf_stuff(resource_t *res, char **env, int depth)
 		return;
 	snprintf(val, n, "%s=%s", OCF_CHECK_LEVEL_STR, ver);
 	*env = val; env++;
+
+	/*
+	   Store the resource local refcnt (0 for now)
+	 */
+	snprintf(ver, sizeof(ver), "%d", refcnt);
+	n = strlen(OCF_REFCNT_STR) + strlen(ver) + 2;
+	val = malloc(n);
+	if (!val)
+		return;
+	snprintf(val, n, "%s=%s", OCF_REFCNT_STR, ver);
+	*env = val; env++;
 }
 
 
@@ -226,7 +217,7 @@ add_ocf_stuff(resource_t *res, char **env, int depth)
    @see			kill_env res_exec add_ocf_stuff
  */
 static char **
-build_env(resource_node_t *node, int depth)
+build_env(resource_node_t *node, int depth, int refcnt)
 {
 	resource_t *res = node->rn_resource;
 	char **env;
@@ -234,7 +225,7 @@ build_env(resource_node_t *node, int depth)
 	int x, attrs, n;
 
 	for (attrs = 0; res->r_attrs && res->r_attrs[attrs].ra_name; attrs++);
-	attrs += 7; /*
+	attrs += 8; /*
 		   Leave space for:
 		   OCF_RA_VERSION_MAJOR
 		   OCF_RA_VERSION_MINOR
@@ -242,6 +233,7 @@ build_env(resource_node_t *node, int depth)
 		   OCF_RESOURCE_INSTANCE
 		   OCF_RESOURCE_TYPE
 		   OCF_CHECK_LEVEL
+		   OCF_RESKEY_RGMANAGER_meta_refcnt
 		   (null terminator)
 		 */
 
@@ -281,7 +273,7 @@ build_env(resource_node_t *node, int depth)
 		++attrs;
 	}
 
-	add_ocf_stuff(res, &env[attrs], depth);
+	add_ocf_stuff(res, &env[attrs], depth, refcnt);
 
 	return env;
 }
@@ -321,6 +313,26 @@ restore_signals(void)
 }
 
 
+/** Find the index for a given operation / depth in a resource node */
+int
+res_act_index(resource_node_t *node, const char *op_str, int depth)
+{
+	int x = 0;
+	resource_act_t *act;
+
+	for (x = 0; node->rn_actions[x].ra_name; x++) {
+		act = &node->rn_actions[x];
+		if (depth != act->ra_depth)
+			continue;
+		if (strcasecmp(act->ra_name, op_str))
+			continue;
+		return x;
+	}
+
+	return -1;
+}
+
+
 /**
    Execute a resource-specific agent for a resource node in the tree.
 
@@ -335,6 +347,8 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 {
 	int childpid, pid;
 	int ret = 0;
+	int act_index;
+	time_t sleeptime = 0;
 	char **env = NULL;
 	resource_t *res = node->rn_resource;
 	const char *op_str = agent_op_str(op);
@@ -343,8 +357,19 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 	if (!res->r_rule->rr_agent)
 		return 0;
 
+	/* Get the action index for later */
+	act_index = res_act_index(node, op_str, depth);
+
+	/* This shouldn't happen, but execing an action for which 
+	   we have an incorrect depth or no status action does not
+	   indicate a problem.  This allows people writing resource
+	   agents to write agents which have no status/monitor function
+	   at their option, in violation of the OCF RA API specification. */
+	if (act_index < 0)
+		return 0;
+
 #ifdef DEBUG
-	env = build_env(node, depth);
+	env = build_env(node, depth, node->rn_resource->r_incarnations);
 	if (!env)
 		return -errno;
 #endif
@@ -373,7 +398,7 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 #endif
 
 #ifndef DEBUG
-		env = build_env(node, depth);
+		env = build_env(node, depth, node->rn_resource->r_incarnations);
 #endif
 
 		if (!env)
@@ -398,11 +423,53 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 	kill_env(env);
 #endif
 
-	do {
-		pid = waitpid(childpid, &ret, 0);
-		if ((pid < 0) && (errno == EINTR))
-			continue;
-	} while (0);
+	if (node->rn_flags & RF_ENFORCE_TIMEOUTS)
+		sleeptime = node->rn_actions[act_index].ra_timeout;
+
+	if (sleeptime > 0) {
+
+		/* There's a better way to do this, but this is easy and
+		   doesn't introduce signal woes */
+		while (sleeptime) {
+			pid = waitpid(childpid, &ret, WNOHANG);
+
+			if (pid == childpid)
+				break;
+			sleep(1);
+			--sleeptime;
+		}
+
+		if (pid != childpid && sleeptime == 0) {
+
+			clulog(LOG_ERR,
+			       "%s on %s:%s timed out after %d seconds\n",
+			       op_str, res->r_rule->rr_type,
+			       res->r_attrs->ra_value,
+			       node->rn_actions[act_index].ra_timeout,
+			       ocf_strerror(ret));
+			
+			/* This can't be guaranteed to kill even the child
+			   process if the child is in disk-wait :( */
+			kill(childpid, SIGKILL);
+			sleep(1);
+			pid = waitpid(childpid, &ret, WNOHANG);
+			if (pid == 0) {
+				clulog(LOG_ERR,
+				       "Task %s PID %d did not exit "
+				       "after SIGKILL\n",
+				       op_str, childpid);
+			}
+
+			/* Always an error if we time out */
+			return 1;
+		}
+	} else {
+		do {
+			pid = waitpid(childpid, &ret, 0);
+			if ((pid < 0) && (errno == EINTR))
+				continue;
+		} while (0);
+	}
 
 	if (WIFEXITED(ret)) {
 
@@ -432,6 +499,39 @@ res_exec(resource_node_t *node, int op, const char *arg, int depth)
 }
 
 
+static inline void
+assign_restart_policy(resource_t *curres, resource_node_t *parent,
+		      resource_node_t *node)
+{
+	char *val;
+	int max_restarts = 0;
+	time_t restart_expire_time = 0;
+
+	node->rn_restart_counter = NULL;
+
+	if (!curres || !node)
+		return;
+	if (parent) /* Non-parents don't get one for now */
+		return;
+
+	val = res_attr_value(curres, "max_restarts");
+	if (!val)
+		return;
+	max_restarts = atoi(val);
+	if (max_restarts <= 0)
+		return;
+	val = res_attr_value(curres, "restart_expire_time");
+	if (val) {
+		restart_expire_time = (time_t)expand_time(val);
+		if ((int64_t)restart_expire_time < 0)
+			return;
+	}
+
+	node->rn_restart_counter = restart_init(restart_expire_time,
+						max_restarts);
+}
+
+
 static inline int
 do_load_resource(int ccsfd, char *base,
 	         resource_rule_t *rule,
@@ -514,6 +614,7 @@ do_load_resource(int ccsfd, char *base,
 	node->rn_state = RES_STOPPED;
 	node->rn_flags = 0;
 	node->rn_actions = (resource_act_t *)act_dup(curres->r_actions);
+	assign_restart_policy(curres, parent, node);
 
 	snprintf(tok, sizeof(tok), "%s/@__independent_subtree", base);
 #ifndef NO_CCS
@@ -526,6 +627,16 @@ do_load_resource(int ccsfd, char *base,
 		free(ref);
 	}
 
+	snprintf(tok, sizeof(tok), "%s/@__enforce_timeouts", base);
+#ifndef NO_CCS
+	if (ccs_get(ccsfd, tok, &ref) == 0) {
+#else
+	if (conf_get(tok, &ref) == 0) {
+#endif
+		if (atoi(ref) > 0 || strcasecmp(ref, "yes") == 0)
+			node->rn_flags |= RF_ENFORCE_TIMEOUTS;
+		free(ref);
+	}
 
 	curres->r_refs++;
 
@@ -666,8 +777,10 @@ build_tree(int ccsfd, resource_node_t **tree,
 			}
 		}
 		/* No resource rule matching the child?  Press on... */
-		if (!flags)
+		if (!flags) {
+			free(ref);
 			continue;
+		}
 
 		flags = 0;
 		/* Don't descend on anything we should have already picked
@@ -687,11 +800,9 @@ build_tree(int ccsfd, resource_node_t **tree,
 			break;
 		}
 
-		if (flags == 2) {
-			free(ref);
-			continue;
-		}
 		free(ref);
+		if (flags == 2)
+			continue;
 
 		x = 1;
 		switch(do_load_resource(ccsfd, tok, childrule, tree,
@@ -769,6 +880,11 @@ destroy_resource_tree(resource_node_t **tree)
 			destroy_resource_tree(&(*tree)->rn_child);
 
 		list_remove(tree, node);
+
+		if (node->rn_restart_counter) {
+			restart_cleanup(node->rn_restart_counter);
+		}
+
 		if(node->rn_actions){
 			free(node->rn_actions);
 		}
@@ -782,7 +898,6 @@ _print_resource_tree(resource_node_t **tree, int level)
 {
 	resource_node_t *node;
 	int x, y;
-	char *val;
 
 	list_do(tree, node) {
 		for (x = 0; x < level; x++)
@@ -799,6 +914,8 @@ _print_resource_tree(resource_node_t **tree, int level)
 				printf("COMMON ");
 			if (node->rn_flags & RF_INDEPENDENT)
 				printf("INDEPENDENT ");
+			if (node->rn_flags & RF_ENFORCE_TIMEOUTS)
+				printf("ENFORCE-TIMEOUTS ");
 			printf("]");
 		}
 		printf(" {\n");
@@ -807,14 +924,11 @@ _print_resource_tree(resource_node_t **tree, int level)
 		     node->rn_resource->r_attrs[x].ra_value; x++) {
 			for (y = 0; y < level+1; y++)
 				printf("  ");
-
-			val = attr_value(node,
-					 node->rn_resource->r_attrs[x].ra_name);
-			if (!val &&
-			    node->rn_resource->r_attrs[x].ra_flags&RA_INHERIT)
-				continue;
 			printf("%s = \"%s\";\n",
-			       node->rn_resource->r_attrs[x].ra_name, val);
+			       node->rn_resource->r_attrs[x].ra_name,
+			       attr_value(node,
+					  node->rn_resource->r_attrs[x].ra_name)
+			      );
 		}
 
 		_print_resource_tree(&node->rn_child, level + 1);
@@ -959,20 +1073,17 @@ _res_op_by_level(resource_node_t **tree, resource_t *first, void *ret,
 		return _res_op(&node->rn_child, first, NULL, ret, op);
 
 	if (op == RS_START || op == RS_STATUS) {
-		rv =  _do_child_levels(tree, first, ret, op);
+		rv |= _do_child_levels(tree, first, ret, op);
 	       	if (rv & SFL_FAILURE)
 			return rv;
 
 		/* Start default level after specified ones */
-		rv =  _do_child_default_level(tree, first, ret, op);
+		rv |= _do_child_default_level(tree, first, ret, op);
 
 	} /* stop */ else {
 
-		rv =  _do_child_default_level(tree, first, ret, op);
-	       	if (rv != 0)
-			return rv;
-
-		rv =  _do_child_levels(tree, first, ret, op);
+		rv |= _do_child_default_level(tree, first, ret, op);
+		rv |= _do_child_levels(tree, first, ret, op);
 	}
 
 	return rv;
@@ -1031,7 +1142,8 @@ do_status(resource_node_t *node)
 
 		/* Ok, it's a 'status' action. See if enough time has
 		   elapsed for a given type of status action */
-		if (delta < node->rn_actions[x].ra_interval)
+		if (delta < node->rn_actions[x].ra_interval ||
+		    !node->rn_actions[x].ra_interval)
 			continue;
 
 		if (idx == -1 ||
@@ -1040,33 +1152,21 @@ do_status(resource_node_t *node)
 	}
 
 	/* No check levels ready at the moment. */
-	if (idx == -1)
+	if (idx == -1) {
+		if (node->rn_checked)
+			return node->rn_last_status;
 		return 0;
+	}
+
 
- 	/* Clear all check levels lower than us */
- 	for (x = 0; node->rn_actions[x].ra_name; x++) {
- 		if (x == idx) {
- 			node->rn_actions[idx].ra_last = now;
- 			continue;
- 		}
- 		if (strcmp(node->rn_actions[x].ra_name, "status"))
- 			continue;
- 
- 		if (node->rn_actions[x].ra_depth <
- 		    node->rn_actions[idx].ra_depth)
- 			node->rn_actions[x].ra_last = now;
- 	}
- 
- 	/*printf("-> %s:%s %s level %d interval = %d\n",
- 		node->rn_resource->r_rule->rr_type,
- 		node->rn_resource->r_attrs->ra_value,
- 		node->rn_actions[idx].ra_name,
- 		node->rn_actions[idx].ra_depth,
- 		(int)node->rn_actions[idx].ra_interval);*/
- 
 	node->rn_actions[idx].ra_last = now;
-	if ((x = res_exec(node, RS_STATUS, NULL,
-                         node->rn_actions[idx].ra_depth)) == 0)
+	x = res_exec(node, RS_STATUS, NULL, node->rn_actions[idx].ra_depth);
+
+	node->rn_last_status = x;
+	node->rn_last_depth = node->rn_actions[idx].ra_depth;
+	node->rn_checked = 1;
+
+	if (x == 0)
 		return 0;
 
 	if (!has_recover)
@@ -1127,14 +1227,18 @@ clear_checks(resource_node_t *node)
 
 	now = res->r_started;
 
-       for (; node->rn_actions[x].ra_name; x++) {
+	for (; node->rn_actions[x].ra_name; x++) {
 
-               if (strcmp(node->rn_actions[x].ra_name, "monitor") &&
-                   strcmp(node->rn_actions[x].ra_name, "status"))
+		if (strcmp(node->rn_actions[x].ra_name, "monitor") &&
+		    strcmp(node->rn_actions[x].ra_name, "status"))
 			continue;
 
-               node->rn_actions[x].ra_last = now;
+		node->rn_actions[x].ra_last = now;
 	}
+
+	node->rn_checked = 0;
+	node->rn_last_status = 0;
+	node->rn_last_depth = 0;
 }
 
 
@@ -1153,11 +1257,11 @@ clear_checks(resource_node_t *node)
    @param realop	Operation to perform if either first is found,
    			or no first is declared (in which case, all nodes
 			in the subtree).
-   @param node		Node we're operating on
    @see			_res_op_by_level res_exec
  */
 static inline int
-_res_op_internal(resource_node_t **tree, resource_t *first,
+_res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
+		 resource_t *first,
 		 char *type, void *__attribute__((unused))ret, int realop,
 		 resource_node_t *node)
 {
@@ -1213,9 +1317,15 @@ _res_op_internal(resource_node_t **tree, resource_t *first,
 
 	/* Start starts before children */
 	if (me && (op == RS_START)) {
-		node->rn_flags &= ~RF_NEEDSTART;
 
-		rv = res_exec(node, op, NULL, 0);
+		if (node->rn_flags & RF_RECONFIG &&
+		    realop == RS_CONDSTART) {
+			rv = res_exec(node, RS_RECONFIG, NULL, 0);
+			op = realop; /* reset to CONDSTART */
+		} else {
+			rv = res_exec(node, op, NULL, 0);
+		}
+		node->rn_flags &= ~(RF_NEEDSTART | RF_RECONFIG);
 		if (rv != 0) {
 			node->rn_state = RES_FAILED;
 			return SFL_FAILURE;
@@ -1253,27 +1363,27 @@ _res_op_internal(resource_node_t **tree, resource_t *first,
 
 	}
 
-	if (node->rn_child) {
-		rv = _res_op_by_level(&node, me?NULL:first, ret, op);
-		if (rv != 0) {
+       if (node->rn_child) {
+                rv |= _res_op_by_level(&node, me?NULL:first, ret, op);
+
+               /* If one or more child resources are failed and at least one
+		  of them is not an independent subtree then let's check if
+		  if we are an independent subtree.  If so, mark ourself
+		  and all our children as failed and return a flag stating
+		  that this section is recoverable apart from siblings in
+		  the resource tree. */
+		if (op == RS_STATUS && (rv & SFL_FAILURE) &&
+		    (node->rn_flags & RF_INDEPENDENT)) {
 			mark_nodes(node, RES_FAILED,
 				   RF_NEEDSTART | RF_NEEDSTOP);
-
-			/* If this node is independent of its siblings,
-			   that one of its dependent children failed
-			   does not matter: its dependent children must
-			   also be independent of this node's siblings. */
-			if (node->rn_flags & RF_INDEPENDENT)
-				return SFL_RECOVERABLE;
-
-			return SFL_FAILURE;
+			rv = SFL_RECOVERABLE;
 		}
 	}
-
+ 			
 	/* Stop should occur after children have stopped */
 	if (me && (op == RS_STOP)) {
 		node->rn_flags &= ~RF_NEEDSTOP;
-		rv = res_exec(node, op, NULL, 0);
+		rv |= res_exec(node, op, NULL, 0);
 
 		if (rv != 0) {
 			node->rn_state = RES_FAILED;
@@ -1435,6 +1545,7 @@ int
 resource_delta(resource_t **leftres, resource_t **rightres)
 {
 	resource_t *lc, *rc;
+	int ret;
 
 	list_do(leftres, lc) {
 		rc = find_resource_by_ref(rightres, lc->r_rule->rr_type,
@@ -1447,10 +1558,25 @@ resource_delta(resource_t **leftres, resource_t **rightres)
 		}
 
 		/* Ok, see if the resource is the same */
-		if (rescmp(lc, rc) == 0) {
+		ret = rescmp(lc, rc);
+		if (ret	== 0) {
 			rc->r_flags |= RF_COMMON;
 			continue;
 		}
+
+		if (ret == 2) {
+			/* return of 2 from rescmp means
+			   the two resources differ only 
+			   by reconfigurable bits */
+			/* Do nothing on condstop phase;
+			   do a "reconfig" instead of 
+			   "start" on conststart phase */
+			rc->r_flags |= RF_COMMON;
+			rc->r_flags |= RF_NEEDSTART;
+			rc->r_flags |= RF_RECONFIG;
+			continue;
+		}
+
 		rc->r_flags |= RF_COMMON;
 
 		/* Resource has changed.  Flag it. */
@@ -1512,12 +1638,17 @@ resource_tree_delta(resource_node_t **ltree, resource_node_t **rtree)
 			   or is new), then we don't really care about its
 			   children.
 			 */
+
 			if (rn->rn_resource->r_flags & RF_NEEDSTART) {
 				rn->rn_flags |= RF_NEEDSTART;
-				continue;
+				if ((rn->rn_resource->r_flags & RF_RECONFIG) == 0)
+					continue;
 			}
 
-			if (rc == 0) {
+			if (rc == 0 || rc == 2) {
+				if (rc == 2)
+					rn->rn_flags |= RF_NEEDSTART | RF_RECONFIG;
+
 				/* Ok, same resource.  Recurse. */
 				ln->rn_flags |= RF_COMMON;
 				rn->rn_flags |= RF_COMMON;
