@@ -326,7 +326,7 @@ add_sha1()
 
 verify_sha1()
 {
-	declare tmpfile="/tmp/smb-$OCF_RESKEY_name.tmp.$$"
+	declare tmpfile="$(mktemp /tmp/smb-$OCF_RESKEY_name.tmp.XXXXXX)"
 	declare current exp
 
 	exp=$(grep "^# rgmanager-sha1.*$1" "$1" | head -1)
