@@ -1,22 +1,3 @@
-/*
-  Copyright Red Hat, Inc. 2002
-  Copyright Mission Critical Linux, 2000
-
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2, or (at your option) any
-  later version.
-
-  This program is distributed in the hope that it will be useful, but
-  WITHOUT ANY WARRANTY; without even the implied warranty of
-  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-  General Public License for more details.
-
-  You should have received a copy of the GNU General Public License
-  along with this program; see the file COPYING.  If not, write to the
-  Free Software Foundation, Inc.,  675 Mass Ave, Cambridge, 
-  MA 02139, USA.
-*/
 /** @file
  * Utility for logging arbitrary strings to the cluster log file via syslog.
  *
@@ -53,10 +34,10 @@ configure_logging(int ccsfd)
 	char *v;
 	char internal = 0;
 
-	if (ccsfd == -1) {
+	if (ccsfd < 0) {
 		internal = 1;
 		ccsfd = ccs_connect();
-		if (ccsfd == -1)
+		if (ccsfd < 0)
 			return -1;
 	}
 
@@ -83,7 +64,7 @@ main(int argc, char **argv)
     int  opt;
     int  severity = 7,
 	 cmdline_loglevel = 0;/* set if we should not use the config file val*/
-    char *logmsg;
+    char *logmsg, *msg_ptr = argv[argc-1];
     int  pid = 0;
     char *progname = NULL;
     int result;
@@ -92,8 +73,7 @@ main(int argc, char **argv)
     if (argc < 4)
 	usage(argv[0]);
 
-    /* ../daemons/log.c */
-    configure_logging(-1);
+    --argc; /* don't parse last value in getopt */
 
     while ((opt = getopt(argc, argv, "f:l:s:hp:n:")) != -1) {
 	switch (opt) {
@@ -121,15 +101,22 @@ main(int argc, char **argv)
 
     /* Add two bytes for linefeed and NULL terminator */
     len = strlen(argv[argc-1]) + 2;
-    logmsg = (char*)malloc(strlen(argv[argc-1])+2);
+    logmsg = (char*)malloc(strlen(msg_ptr)+2);
+    if (logmsg == NULL) {
+        fprintf(stderr,
+            "clulog: malloc fail err=%d\n", errno);
+        exit(0);
+    }
+
     snprintf(logmsg, len, "%s\n", argv[argc-1]);
 
     if (!cmdline_loglevel) {
 	/*
 	 * Let's see what loglevel the SM is running at.
-	 * TODO Get rgmgr log level
+	 * If ccsd's not available, use default.
 	 */
-	clu_set_loglevel(LOGLEVEL_DFLT);
+    	if (configure_logging(-1) < 0)
+		clu_set_loglevel(LOGLEVEL_DFLT);
     }
     result = clulog_pid(severity, pid, progname, logmsg);
     free(progname);
